/*
 * Copyright (c) 1996 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Computer Science
 *      Department at University College London
 * 4. Neither the name of the University nor of the Department may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "prototypes.h"
#include <stdio.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>

#define NEW_SOLUTION
extern block *find_block_by_id();
extern line *find_line_by_id();
extern line *find_last_line();
/*extern block *init_block();*/
extern line *init_line();
extern void redisplay();
extern typeface tf;
extern page *p;
extern Tcl_Interp *interp;
extern u_int16 lkey;
extern struct in_addr ipaddr;
/*#define DEBUG*/
/*#define ABORT*/

/*Seen page is set to TRUE when we see a page message from someone else,
  or get requested a page message from someone else.
  It's used to facilitate late joining*/
int seen_page=FALSE;

/********************************************************************/
/*parse_line: called when we have a LINEMSG received from the net   */
/*            this fn decides what to do with the received line     */
/********************************************************************/
int parse_line(nl, t)
     netline *nl;
     struct timeval *t;
{
  block *b;
  line *l;
  char command[80];
  char blockid[ID_LEN];
  char lineid[ID_LEN];
  char previd[ID_LEN];
  char nextid[ID_LEN];
  user_data modifier;
  int no_of_lines;

  ntohid(&(nl->blockid), blockid);
  ntohid(&(nl->lineid), lineid);
  ntohuser(&(nl->modifier), &modifier);

  if(ntohs(nl->linenum)==UNKNOWN_LINE)
    {
      /*The site sending a sliding key is telling me they've never heard 
	of this line.  Remove it from the missing list so I don't ask
	them again, and only add it back if I see it referenced by
	someone else.

	to do this properly we should keep track of who tells me they haven't
	seen it, but getting told once will do for now.*/
      if (is_in_missing_list(lineid))
	rm_from_missing_list(lineid);
      return 0;
    }
#ifdef DEBUG
  printf("--------\ngot line with block id: %s\n", blockid);
  printf("modifier: >%s<\n", modifier.username);
#endif
  b=find_block_by_id(p, blockid);
  if(b==NULL)
    {
      /*this block is a new one on me...*/
      /*we'll have to create it now, and fill in the details later*/
      block *prev;
      block *new_block;
      user_data originator;
      set_null_name(&originator);
#ifdef DEBUG
      printf("New line in new block recv'ed\n");
#endif
      prev=p->last_block;
      
      add_to_missing_list(blockid, BLOCKMSG);
      new_block=init_block(blockid, 
			   NORMAL,
			   nl->xpos,
			   ntohs(nl->ypos),
			   t,
			   prev,
			   NULL,
			   &originator,
			   &(modifier),
			   &tf);
      set_null_name(&originator);
      sprintf(command, "set_block_posn %s %d %d", blockid, -1, -1);
      Tcl_Eval(interp, command);
      if(p->first_block==NULL)
	p->first_block=new_block;
      p->last_block=new_block;
      b=new_block;
    }

  /*OK, now we have a block to work with...*/
  ntohid(&(nl->previd), previd);
  ntohid(&(nl->nextid), nextid);
#ifdef DEBUG
  printf("prev line: %s\n", previd);
  printf("next line: %s\n", nextid);
#endif
  l=find_line_by_id(lineid, b);
  if(l==NULL)
    {
      line *prevl;
      line *nextl;
#ifdef DEBUG
      printf("Got a new line with id %s\n", lineid);
#endif
      /*this is a line we haven't seen before*/
      if(strncmp(lineid,"NULL",ID_LEN)!=0)
	if(is_in_missing_list(lineid))
	  rm_from_missing_list(lineid);

      prevl=find_line_by_id(previd, b);
      if(prevl==NULL)
	{
	  if (strncmp(previd,"NULL",ID_LEN)!=0)
	    {
	      add_to_missing_list(previd, LINEMSG);
#ifdef DEBUG
	      printf("*haven't seen the previous line\n");
#endif
	    }
	  /*haven't seen the previous line either*/
	  nextl=find_line_by_id(nextid, b);
	  if(nextl==NULL)
	    {
	      if (strncmp(nextid,"NULL",ID_LEN)!=0)
		{
		  add_to_missing_list(nextid, LINEMSG);
#ifdef DEBUG
		  printf("*haven't seen the next line either\n");
#endif
		}
	      /*OK, haven't seen either of the neightbouring lines*/
	      /*tag it onto the end of the list and worry about it later*/
	      find_best_place_for_line(b, ntohs(nl->linenum), &prevl, &nextl);
	    }
	  else
	    {
	      prevl=nextl->prev_line;
	    }
	}
      else
	{
	  nextl=prevl->next_line;
	  if(nextl==NULL)
	    {
#ifdef DEBUG
		  printf("*haven't seen the next line\n");
#endif
	      add_to_missing_list(nextid, LINEMSG);
	    }
	  else if (strncmp(nextid, nextl->lineid, ID_LEN)!=0)
	    {
	      add_to_missing_list(nextid, LINEMSG);
	    }
	}
	
      l=init_line(lineid, ntohs(nl->linenum), t, prevl, previd, nextl, nextid);
      add_to_recent_list(lineid, t, LINEMSG);

      if((prevl!=NULL)&&(strcmp(l->lineid, prevl->nextid)!=0))
	{
#ifdef DEBUG
	  printf("l->linenum: %d\nprev->linenum: %d\n", l->linenum, prevl->linenum);
#endif
	  if((l->linenum-prevl->linenum)>0)
	    l->missing_lines=(l->linenum-prevl->linenum)-1;
	  else
	    l->missing_lines=0;
	}
      else
	l->missing_lines=0;

      if((nextl!=NULL)&&(strcmp(l->nextid, nextl->lineid)!=0)
	 &&(nextl->missing_lines>0))
	{
#ifdef DEBUG
	  printf("still got some missing lines after this one...");
#endif
	  if((nextl->linenum-ntohs(nl->linenum))>0)
	    nextl->missing_lines=(nextl->linenum-ntohs(nl->linenum))-1;
	  else
	    nextl->missing_lines=0;
	}
      else
	if(nextl!=NULL) 
	  {
#ifdef DEBUG
          printf("no missing lines after this one...");
#endif
	    nextl->missing_lines=0;
	  }
#ifdef DEBUG
      else
	{
          printf("no lines after this one...");
	}
#endif



      if (nl->no_of_chars>0)
	strncpy(l->line_data, nl->line_data, MAX_LINE_LEN);
      else
	strcpy(l->line_data, " ");

      l->no_of_chars=nl->no_of_chars;
      l->block=b;
      if(newer(t, &(b->last_mod)))
	time_copy(t, &(b->last_mod));
      if(newer(t, &(p->last_mod)))
	 time_copy(t, &(p->last_mod));
      user_copy(&modifier, &(b->modifier));

      if((b->first_line==NULL)||(prevl==NULL))
	b->first_line=l;

      no_of_lines=count_lines(b);
      if(no_of_lines > b->no_of_lines) b->no_of_lines=no_of_lines;

      if((b->xpos!=nl->xpos)||(b->ypos!=ntohs(nl->ypos)))
	{
	  /*block has moved*/
	  b->xpos=nl->xpos;
	  b->ypos=ntohs(nl->ypos);
	  /*need to redisplay whole block now*/
	}

      if(prevl==NULL)
	{
	  u_int16 lnum;
	  if(l->prev_line==NULL) lnum=0;
	  else
	    lnum=find_line_num(b, l->prev_line->lineid);
	  l->missing_lines=l->linenum-lnum;
	  /*now need to get those missing lines*/
	}

      else if(ntohs(nl->linenum)!=find_line_num(b, l->lineid))
	{
	  /*someone inserted or deleted a line before this one,
	    and we missed the change*/
	  /*need to request enough info to figure out where the
	    line was inserted or deleted*/
	}


      /*fix up the line numbers so we don't get too confused*/
      renumber_block(b);

      /*now need to force the update on screen*/
#ifdef DEBUG
      printf("new line: ->%s<-\n", l->line_data);
#endif
      if(b->status!=DELETED) delayed_redisplay_block(b);
    }
  else
    {
      /*we've seen the line before*/
#ifdef DEBUG
      printf("got a modified line\n");
#endif
      if(newer(t, &(l->last_mod)))
	{
	  line *nextl;
	  line *prevl;
#ifdef DEBUG
	  printf("nt=%ld,%ld  ot=%ld,%ld\n", t->tv_sec, t->tv_usec, (l->last_mod).tv_sec, (l->last_mod).tv_usec);
#endif
	  /*this is a later version*/
	  if(is_in_missing_list(lineid))
	    rm_from_missing_list(lineid);
	  strncpy(l->line_data, nl->line_data, MAX_LINE_LEN);
#ifdef DEBUG
	  printf("data:%s\n", l->line_data);
#endif
	  l->no_of_chars=nl->no_of_chars;
#ifdef DEBUG
	  printf("line has %d (%d) chars\n", l->no_of_chars, strlen(l->line_data));
#endif
	  time_copy(t, &(l->last_mod));
	  time_copy(t, &(b->last_mod));
	  time_copy(t, &(p->last_mod));
	  user_copy(&(modifier), &(l->modifier));
	  user_copy(&(modifier), &(b->modifier));

#ifdef DEBUG
	  printf("2: prev line: %s\n", previd);
	  printf("2: next line: %s\n", nextid);
#endif
	  /*new lines may have been inserted before or after this one,
	    but we haven't received the new line yet...*/
	  strcpy(l->nextid, nextid);
	  strcpy(l->previd, previd);
	  
	  nextl=l->next_line;
	  prevl=l->prev_line;
	  /*did we just get informed there should be a line before this
	    one that we haven't seen yet?*/
	  if ((prevl!=NULL)&&(strcmp(prevl->lineid,previd)!=0))
	    if((l->linenum - prevl->linenum)>0)
	      l->missing_lines=(l->linenum - prevl->linenum)-1;

	  /*unfortunately if we got informed there should be a line after
	    this one and haven't seen it yet, we also don't know now
	    many lines are missing - so assume we're only missing 1*/
	  if ((nextl!=NULL)&&(strcmp(nextl->lineid,nextid)!=0))
	    if (nextl->missing_lines==0)
	      nextl->missing_lines=1;
	    

#ifdef DEBUG
	  if (nextl!=NULL)
	    {
	      printf("nextl->missing_lines: %d\n", nextl->missing_lines);
	      printf("nextl->lineid: %s\n", nextl->lineid);
	    }
	  printf("l->nextid: %s\n", l->nextid);
#endif
	  if ((nextl!=NULL)&&(nextl->missing_lines!=0)
	      &&((strncmp(l->nextid, nextl->lineid, ID_LEN)==0)
		 ||(strncmp(l->nextid, "OK", ID_LEN)==0)))
	    {
#ifdef DEBUG
	      printf("zeroing nextl->missing_lines\n");
#endif
	      nextl->missing_lines=0;
	    }
	  add_to_recent_list(lineid, t, LINEMSG);
	  if((b->xpos!=nl->xpos)||(b->ypos!=ntohs(nl->ypos)))
	    {
	      /*block has moved*/
	      b->xpos=nl->xpos;
	      b->ypos=ntohs(nl->ypos);
	      /*need to redisplay whole block now*/
	    }
	  if(ntohs(nl->linenum)!=find_line_num(b, l->lineid))
	    {
	      /*someone inserted or deleted a line before this one,
		and we missed the change*/
	      /*need to request enough info to figure out where the
	       line was inserted*/
	    }
	  if((ntohs(nl->linenum)==LINE_DELETED)&&(l->linenum!=LINE_DELETED))
	    {
#ifdef DEBUG
	      printf("got a LINE DELETION, >>%s<<\n", lineid);
#endif
	      l->linenum=LINE_DELETED;
	      delete_line(l, b);
	    }

	  /*now need to force the update on screen*/
	  if(b->status!=DELETED) redisplay_block(b);
	}
      else
	{
#ifdef DEBUG
	  printf("must be an old packet - ignoring it\n");
	  printf("nt=%ld,%ld  ot=%ld,%ld\n", t->tv_sec, t->tv_usec, l->last_mod.tv_sec, l->last_mod.tv_usec);
	  printf("data: >>%s<<\n", l->line_data);
	  printf("id: %s\n", l->lineid);
#endif
	  if(is_in_missing_list(lineid))
	    rm_from_missing_list(lineid);
	}
    }
  return 0;
}


/********************************************************************/
/*parse_block: called when we have a BLOCKMSG received from the net */
/*            this fn decides what to do with the received block    */
/********************************************************************/
int parse_block(nb, t)
     netblock *nb;
     struct timeval *t;
{
  block *b;
  char command[80];
  char blockid[ID_LEN];
  char previd[ID_LEN];
  char nextid[ID_LEN];
  user_data originator;
  user_data modifier;
  ntohid(&(nb->blockid), blockid);
  ntohid(&(nb->previd), previd);
  ntohid(&(nb->nextid), nextid);
  ntohuser(&(nb->originator), &originator);
  ntohuser(&(nb->modifier), &modifier);

#ifdef DEBUG
      printf("got a block with id %s\n", blockid);
#endif
  b=find_block_by_id(p, blockid);
  if(b==NULL)
    {
      /*This is a new block*/
      block *new_block;
      block *prev=NULL;
      block *next=NULL;
#ifdef DEBUG
      printf("got a new block\n");
#endif
      if(is_in_missing_list(blockid))
	rm_from_missing_list(blockid);
#ifdef OLD_SOLUTION
      if(strncmp(previd, "NULL")!=0)
        prev=find_block_by_id(p, previd);
      if(prev==NULL)
	{
	  /*we don't have the previous block either!*/
	  if(strncmp(nextid, "NULL")!=0)
	    next=find_block_by_id(p, nextid);
	  if(next==NULL)
	    {
	      /*we don't have the next block either!*/
	      /*OK, we just received a block, and we haven;t seen either
		of the blocks that are supposed to neighbour it*/
	      /*we'll tag it on the end of our blocklist and hope the
		neighbours turn up*/
	      /*eventually we'll have to request their re-transmission*/
	      if(p->last_block==NULL)
		{
		  /*This is the first block we've ever seen*/
		  new_block=init_block(blockid, 
				       NORMAL,
				       nb->xpos,
				       ntohs(nb->ypos),
				       t,
				       NULL,
				       NULL,
				       &(originator),
				       &(modifier),
				       &(nb->face));
		  p->first_block=new_block;
		  p->last_block=new_block;
		  sprintf(command, "set_block_posn %s %d %d", 
			  blockid, nb->xpos, ntohs(nb->ypos));
		  Tcl_Eval(interp, command);
		  sprintf(command, "set_block_colour %s %d", 
			  blockid, nb->face.colindex);
		  Tcl_Eval(interp, command);
		  sprintf(command, "set_block_font %s %d", 
			  blockid, nb->face.fontindex);
		  Tcl_Eval(interp, command);
		  add_to_recent_list(blockid, t, BLOCKMSG);
		  return 0;
		}
	      else
		{
		  prev=p->last_block;
		  next=NULL;
		}
	    }
	  else
	    {
	      /*we had seen the next block but not the previous*/
	      prev=next->prev_block;
	    }
	} 
      else
	{
	  /*we had seen the previous block*/
	  next=prev->next_block;
	}
#endif
#ifdef NEW_SOLUTION
      next=NULL;
      prev=p->last_block;
      if (strncmp(previd, "NULL", ID_LEN)!=0)
	if(find_block_by_id(p, previd)==NULL)
	  if (!is_in_missing_list(previd))
	    add_to_missing_list(previd, BLOCKMSG);
      if (strncmp(nextid, "NULL", ID_LEN)!=0)
	if(find_block_by_id(p, nextid)==NULL)
	  if (!is_in_missing_list(nextid))
	    add_to_missing_list(nextid, BLOCKMSG);
#endif
#ifdef DEBUG
      printf("init block id: %s\n", blockid);
#endif DEBUG
      new_block=init_block(blockid, 
			   nb->status,
			   nb->xpos,
			   ntohs(nb->ypos),
			   t,
			   prev,
			   next,
			   &(originator),
			   &(modifier),
			   &(nb->face));
      if(p->first_block==NULL)
	p->first_block=new_block;
      p->last_block=new_block;
      add_to_recent_list(blockid, t, BLOCKMSG);
      sprintf(command, "set_block_posn %s %d %d", 
	      blockid, nb->xpos, ntohs(nb->ypos));
      Tcl_Eval(interp, command);
      sprintf(command, "set_block_colour %s %d", 
	      blockid, nb->face.colindex);
      Tcl_Eval(interp, command);
      sprintf(command, "set_block_font %s %d", 
	      blockid, nb->face.fontindex);
      Tcl_Eval(interp, command);
#ifdef OLD_SOLUTION
      if(p->last_block==prev)
	{
	  /*we just added a block to the end of the list,
	    so update the end of list pointer in p*/
	  p->last_block=new_block;
	}
#endif
      if(newer(t, &(p->last_mod)))
      	  time_copy(t, &(p->last_mod));

    }
  else
    {
      /*we've seen this block before*/
      if(newer(t, &(b->last_mod)) || is_a_null_name(&(b->originator)))
	{
#ifdef DEBUG
	  printf("got a modified block\n");
#endif
	  if(is_in_missing_list(blockid))
	    rm_from_missing_list(blockid);

	  b->no_of_lines=ntohs(nb->no_of_lines);
	  if((b->xpos != nb->xpos)||(b->ypos != ntohs(nb->ypos)))
	    {
	      b->xpos=nb->xpos;
	      b->ypos=ntohs(nb->ypos);
	      sprintf(command, "set_block_posn %s %d %d", 
		      blockid, nb->xpos, ntohs(nb->ypos));
	      Tcl_Eval(interp, command);
	    }
	  user_copy(&(originator), &(b->originator));
	  user_copy(&(modifier), &(b->modifier));
	  if(b->face.colindex != (nb->face.colindex))
	    {
	      typeface_copy(&(nb->face), &(b->face));
	      sprintf(command, "set_block_colour %s %d", 
		      blockid, nb->face.colindex);
	      Tcl_Eval(interp, command);
	    }
	  if(/*b->face.fontindex != (nb->face.fontindex)*/1)
	    {
#ifdef DEBUG
	      printf("setting block font to %d\n", nb->face.fontindex);
#endif
	      typeface_copy(&(nb->face), &(b->face));
	      sprintf(command, "set_block_font %s %d", 
		      blockid, nb->face.fontindex);
	      Tcl_Eval(interp, command);
	    }
	  if(b->status != nb->status)
	    {
	      switch(nb->status)
		{
		case NORMAL:
		case LOCKED:
		  if (b->status!=DELETED) 
		    b->status=nb->status;
		  else
		    {
		      /*if we see a modification to a deleted block,
			need to ensure someone tells them, but we
			also want anyone else who sees this message
			to set their modification time to the same
			value, and it must be greater than the person
			who thinks this block still exists*/
		      t->tv_usec++;
		      if(t->tv_usec>=1000000)
			{
			  t->tv_sec++;
			  t->tv_usec=0;
			}
		    }
		  break;
		case DELETED:
		  b->status=DELETED;
		  sprintf(command, "delete_block %s", blockid);
		  Tcl_Eval(interp, command);
		}
	    }
	  time_copy(t, &(b->last_mod));
	  time_copy(t, &(p->last_mod));
	  add_to_recent_list(blockid, t, BLOCKMSG);
	}
      else
	{
#ifdef DEBUG
	  printf("got an old block\n");
	  printf("nt=%lu,%lu  ot=%lu,%lu\n", t->tv_sec, t->tv_usec, b->last_mod.tv_sec, b->last_mod.tv_usec);
#endif
	  if((nb->status==DELETED)&(b->status!=DELETED))
	    {
	      b->status=DELETED;
	      sprintf(command, "delete_block %s", blockid);
	      Tcl_Eval(interp, command);
	    }
	}
    }
  return 0;
}

u_int16 masks[17]={0x0, 0x1, 0x3, 0x7, 0xf, 0x1f, 0x3f, 0x7f, 0xff,
0x1ff, 0x3ff, 0x7ff, 0xfff, 0x1fff, 0x3fff, 0x7fff, 0xffff};

int parse_rtx_advert(nrtxa, remote_addr, t)
     netrtxadvert *nrtxa;
     struct in_addr *remote_addr;
     struct timeval *t;
{
  struct timeval last_mod;
  seen_another_sliding_key(t);
  if(((ntohs(nrtxa->key))&(masks[nrtxa->mask]))!=(lkey&(masks[nrtxa->mask])))
     {
       /*key didn't match*/
       return -1;
     }
#ifdef DEBUG
  printf("Key matched...");
#endif
  get_last_mod_time(p, &last_mod);
  if (seen_page==FALSE)
    {
      send_page_request(remote_addr);
      /*you only get to ask once.  if your packet gets lost, tough!*/
      /*I'll change this later*/
      seen_page=TRUE;
    }
  else if (number_missing()>0)
    {
      send_specific_rtx_request(remote_addr);
    }
  else if(newer(t, &last_mod))
    {
      /*we need a retransmission*/
#ifdef DEBUG
      printf("and we need a retransmission (t: %lu,%lu, lm: %lu,%lu\n",
	     t->tv_sec, t->tv_usec, last_mod.tv_sec, last_mod.tv_usec);
#endif
      send_rtx_request(remote_addr, &last_mod);
    }
  return 0;
}

int parse_rtx_request(nrtxr, t)
     netrtxrequest *nrtxr;
     struct timeval *t;
{
#ifdef DEBUG
  printf("Got an RTX REQUEST\n");
#endif
  if((ntohl(nrtxr->source.s_addr))==(ipaddr.s_addr))
    {
#ifdef DEBUG
      printf("The request is addressed to me!\n");
#endif
      send_retransmissions(t);
    }
  return 0;
}

int parse_rtx_specific_request(nrtxr)
     netrtxspecificrequest *nrtxr;
{
  char id[ID_LEN];
#ifdef DEBUG
  printf("Got an RTX SPECIFIC REQUEST\n");
#endif
  if(ntohl(nrtxr->source.s_addr)==(ipaddr.s_addr))
    {
      int i;
#ifdef DEBUG
      printf("The request is addressed to me!\n");
#endif
      for(i=0;i<ntohl(nrtxr->no_of_items);i++)
	{
	  ntohid(&(nrtxr->ids[i]), id);
#ifdef DEBUG
printf("request for %s\n", id);
#endif
	  send_specific_retransmission(id);
	}
    }
  else
    {
/*      printf("It was addressed to %s\n", inet_ntoa(nrtxr->source.s_addr));*/
    }
  return 0;
}

int parse_page_request(netpagerequest *nrtxr)
{
  struct timeval t;
#ifdef DEBUG
  printf("Got an PAGE_REQUEST\n");
#endif
  if(ntohl(nrtxr->source.s_addr)==(ipaddr.s_addr))
    {
#ifdef DEBUG
      printf("The request is addressed to me!\n");
#endif
      t.tv_sec=0;
      t.tv_usec=0;
      send_retransmissions(&t);
    }
  seen_page=TRUE;
  return 0;
}

int parse_session_message(netsession *sm)
{
  struct timeval t;
  struct timezone tzp;
  gettimeofday(&t, &tzp);
  sm->textname[sm->textnamelen]='\0';
  sm->textname[sm->textnamelen+1]='\0';
  add_receiver_to_participants(&(sm->ud), sm->textname, sm->col, &t);
  return 0;
}

int parse_leave_message(netsession *sm)
{
  rm_receiver_from_participants(&(sm->ud), sm->textname);
  return 0;
}

int parse_pointer(netpointer *np)
{
  char str[256];
  char id[ID_LEN];
  ntohid(&(np->id), id);
  if(ntohl(np->x) != DELETE_POINTER)
    {
#ifdef DEBUG
      printf("move_shared_pointer %s %s %d %d %d %s", id, np->username,
              (int)ntohl(np->x), (int)ntohl(np->y), (int)np->col, np->style);
#endif
      sprintf(str, "move_shared_pointer %s %s %d %d %d %s", id, np->username, 
	      (int)ntohl(np->x), (int)ntohl(np->y), (int)np->col, np->style);
      Tcl_Eval(interp, str);
    }
  else
    {
      sprintf(str, "delete_shared_pointer %s", id);
      Tcl_Eval(interp, str);
    }
  return 0;
}

int parse_recent_list(netrecentadvert *nra)
{
  int i;
  char id[ID_LEN];
  struct timeval t;
  block *b;
  line *l;
  for(i=0;i<ntohl(nra->no_of_entries);i++)
    {
      ntohid(&(nra->entries[i].id), id);
      ntohtime(&(nra->entries[i].last_mod), &t);
      if(nra->entries[i].type == BLOCKMSG)
	{
#ifdef DEBUG
	  printf("recent list contains block %s\n", id);
#endif
	  b=find_block_by_id(p, id);
	  if(b==NULL) 
	    {
	      add_to_missing_list(id, BLOCKMSG);
	      continue;
	    }
	  if(newer(&t, &(b->last_mod)))
	    {
#ifdef DEBUG
printf("recent list block more recent than mine: %s\n", id);
printf("recent list time: %lu, %lu\n", t.tv_sec, t.tv_usec);
printf("my block time: %lu, %lu\n", (b->last_mod).tv_sec, (b->last_mod).tv_usec);
#endif
	      add_to_missing_list(id, BLOCKMSG);
	    }
	  else if(newer(&(b->last_mod), &t))
	    {
/*	      add_to_send_list(id, b, BLOCKMSG);*/
	    }
	}
      else if(nra->entries[i].type == LINEMSG)
	{
	  l=NULL;
	  b=p->first_block;
	  while(b!=NULL)
	    {
	      l=find_line_by_id(id, b);
	      if(l==NULL)
		{
		  b=b->next_block;
		}
	      else
		break;
	    }
	  if(l==NULL) 
	    {
	    add_to_missing_list(id, LINEMSG);
	  }
	  else
	    {
	      if(newer(&t, &(l->last_mod)))
		{
#ifdef DEBUG
printf("Need a refresh: %s\n", id);
printf("%ld,%ld older that %ld,%ld\n", l->last_mod.tv_sec, l->last_mod.tv_usec,
t.tv_sec, t.tv_usec);
printf("recent list line more recent than mine: %s\n", id);
printf("recent list time: %lu, %lu\n", t.tv_sec, t.tv_usec);
printf("my line time: %lu, %lu\n", (l->last_mod).tv_sec, (l->last_mod).tv_usec);
#endif
		  add_to_missing_list(id, LINEMSG);
		}
	      else if(newer(&(l->last_mod), &t))
		{
/*		  add_to_send_list(id, l, LINEMSG);*/
		}
	    }
	}
      else
	{
	  ntohid(&(nra->entries[i].id), id);
	  fprintf(stderr, "Got an unknown object type %d in recent advert from %s\n", nra->entries[i].type, id);
	}
    }
  return 0;
}
