/*
 * Copyright (c) 1996 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Computer Science
 *      Department at University College London
 * 4. Neither the name of the University nor of the Department may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <math.h>
#include <stdlib.h>
#include "prototypes.h"
int confsize;

/*sliding key used for retransmissions*/
sliding_key skey;

/*our local key for matching against the sliding key of others*/
u_int16 lkey;

/*default interval between sending tokens in milliseconds*/
/*increase this with each successive sliding key cycle with no mods*/
#define MAX_PASS 4
int inter_token_interval[MAX_PASS]={200,400,800,1600};


/*the file descriptor to send from*/
extern int tx_fd;

extern queue sendq;
extern page *p;

void init_key(sliding_key *skey, u_int16 *lkey)
{
  struct timeval t;
  struct timezone tzp;
  gettimeofday(&t, &tzp);
  lbl_srandom(t.tv_usec);
  confsize=INITIAL_CONF_SIZE;
  skey->master=FALSE;
  skey->mods=FALSE;
  skey->counter=RTX_TOKEN_STOP;
  skey->pass=0;
  *lkey=get_key();
}

u_int16 get_key()
{
  return((u_int16)(lbl_random()&0xffff));
}

void reset_sliding_key(sliding_key *skey, int conf_size)
{
  skey->key=get_key();
  if(conf_size>0)
    skey->mask=(u_int8)((log10((double)conf_size)/log10(2.0))+1.5);
  else
    skey->mask=INITIAL_MASK;
  get_last_mod_time(p, &(skey->timestamp));
  skey->master=TRUE;
  skey->counter=0;
  send_recent_list(tx_fd);
}

void init_rtx_token(sliding_key *skey)
{
  confsize=get_conf_size();
  if(skey->counter==RTX_TOKEN_STOP)
    {
      reset_sliding_key(skey, confsize);
      send_rtx_token();
    }
  else
    {
      skey->mods=TRUE;
      skey->pass=0;
    }
}

void send_rtx_token(/*uses global skey and tx_fd*/)
{
  ntheader header;
  netrtxadvert token;
  if(packets_queued(&sendq)>0)
    {
      /*if we've already got stuff in the queue, don't send this time*/
      /*just reschedule us for the next time*/
      skey.event_token=
	Tcl_CreateTimerHandler(inter_token_interval[skey.pass],
			      (Tcl_TimerProc*)send_rtx_token,
			      (ClientData)0);
      return;
    }
  if(skey.counter++>=TOKENS_PER_MASK_POSN)
    slide_rtx_key(&skey);
  if(skey.counter==RTX_TOKEN_STOP)
    return;
#ifdef DEBUG
  printf("Key %d, mask %d\n", skey.key, skey.mask);
#endif
  init_header(&header, RTX_ADVERT, &(skey.timestamp));
  token.key=htons(skey.key);
  token.mask=skey.mask;
  xsocksend(tx_fd, &header, (u_char *)&token, sizeof(netrtxadvert));
  skey.event_token=
    Tcl_CreateTimerHandler(inter_token_interval[skey.pass],
			  (Tcl_TimerProc*)send_rtx_token,
			  (ClientData)0);
}

void slide_rtx_key(sliding_key *skey)
{
  if (skey->mask==0)
    {
      if((skey->master==TRUE)||(skey->mods==TRUE))
	{
	  /*Once we start a key cycle, we're the master again until
	    we see another sliding key*/
	  skey->master=TRUE;
	  /*Reset the mods flag, until more modifications occur*/
	  if(skey->mods==TRUE)
	    {
	      skey->mods=FALSE;
	      skey->pass=0;
	    }
	  else
	    {
	      if(skey->pass<MAX_PASS-1)
		skey->pass++;
	    }
	  reset_sliding_key(skey, confsize);
	}
      else
	{
	  skey->counter=RTX_TOKEN_STOP;
	}
    }
  else
    {
      skey->mask--;
      skey->counter=0;
      skey->holds=0;
    }
}

void seen_another_sliding_key(struct timeval *t)
{
  struct timeval last_mod;
  if((skey.counter==RTX_TOKEN_STOP)||(skey.mods==TRUE))
    return;
  get_last_mod_time(p, &last_mod);
  if(newer(t, &last_mod))
     skey.master=FALSE;
}

void hold_key()
{
  skey.counter=0;
  skey.pass=0;
  skey.holds++;
  if(skey.holds>MAX_NO_OF_HOLDS) reset_sliding_key(&skey, confsize);
}

